/*
 * Threaded-samba-scanner | GPLv3 or later (C) 2008 - 2011 Magnus Loef
 *
 * 1. Gets IPv4-adresses for all network interfaces.
 * 2. Uses threads to run forked popen calls using
 *    "nmblookup -A IP" for all valid LAN IPs.
 * 3. Shows SAMBA computer names and their IPs.
 *
*/


#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <pthread.h>

#define MAX_INTERFACES 128

/* Scan timeout (9000-999999) */
int global_scan_timeout = 20000;
/* Verbosity level (0 or 1) */
int global_verbose = 0;


char * allocate(size_t alloc_size)
{
    char *ret;

    ret = malloc(alloc_size);
    if( ret == NULL )
    {
        printf("Error: Memory allocation failed or 0 bytes requested.\n");
        exit(-1);
    }
    memset(ret, 0, alloc_size);

    return ret;
}

void show_help()
{
    printf("\n\n--- Threaded-SAMBA-Scanner help ---\n\n");
    printf("Options:\n\n");
    printf("--timeout X : Increase scan timeout. Range is: 9000-999999\n");
    printf("--verbose   : Shows more scanning information.\n");
    printf("--help      : Shows this help.\n\n");
}


int nmblookup_exists()
{
    FILE *fp;
    char *cmd, *line;
    int ret = 0;

    cmd = allocate(16);
    snprintf(cmd, 16, "nmblookup -V");

    if((fp=popen(cmd, "r"))==NULL)
    {
        free(cmd);
        return 0;
    }
    line = allocate(4096);
    while((fgets(line, 4000, fp))!=NULL)
    {
        if( strstr(line, "Version") )
        {
            ret = 1;
            break;
        }
    }
    pclose(fp);
    free(cmd);
    free(line);

    return ret;
}


void * do_lookup(void *input_addr)
{
    FILE *fp;
    char *line, *cmd, *samba_host, *new_addr;
    int begin = 0, end = 0;

    if( global_verbose )
        printf("Looking up address: %s\n", (char *)input_addr);

    if( (char *)input_addr == NULL
    ||  strlen((char *)input_addr) < 3 )
    {
        printf("Error: Recieved bad input address: %s\n", (char *)input_addr);
        return((void *)NULL);
    }

    cmd = allocate(1024);
    snprintf(cmd, 1000, "nmblookup -A %s", (char *)input_addr);

    /* Fork() an nmblookup process for this IP */
    if((fp=popen(cmd, "r"))==NULL)
    {
        printf("Scan command failed: %s\n", cmd);
        free(cmd);
        return((void *)NULL);
    }

    line       = allocate(4096);
    samba_host = allocate(4096);
    new_addr   = allocate(1024);

    /* Since its forked we must get names and
       adresses from the forked() process. */
    while((fgets(line, 4000, fp))!=NULL)
    {
        if( strlen(line) > 24 && strstr(line, "Looking up status of") )
        {
            snprintf(new_addr, 1000, "%s", &line[21]);
            new_addr[strlen(new_addr)-1]='\0';
        }

        if( strstr(line, "<") && strstr(line, ">") && strlen(line) < 4000 )
        {
            for(begin=0; line[begin]!='\0'; begin++)
                if( line[begin]!=' ' && line[begin]!='\t' )
                    break;
            for(end=begin; line[end]!='\0'; end++)
                if( line[end]==' ' )
                    break;
            snprintf(samba_host, end, "%s", &line[begin]);

            printf("Found samba host: [%s] with address: [%s]\n", samba_host, new_addr);
            break;
        }
    }
    pclose(fp);
    free(cmd);
    free(line);
    free(samba_host);
    free(new_addr);

    pthread_exit(NULL);
    return((void *)NULL);
}


void scan_samba_hosts(char *net_addr)
{
    char *ip_addr;
    int ret = 0, i = 0;
    pthread_t threads[254];

    if( net_addr == NULL )
    {
        printf("Error: Invalid network address.\n");
        return;
    }
    printf("Scanning network address range: %s.1-254\n", net_addr);

    ip_addr = allocate(1024);
    /* Scan IPv4 networks (1-254) */
    for(i=1; i < 255; i++)
    {
        memset(ip_addr, 0, 1024);
        snprintf(ip_addr, 1000, "%s.%d", net_addr, i);

        ret = pthread_create(&threads[i-1], NULL, do_lookup, (void *)ip_addr);
        if( ret != 0 )
        {
            printf("Error: Return code from pthread_create is: %d for thread: %d\n", ret, i-1);
            free(net_addr);
            free(ip_addr);
            pthread_exit(NULL);
            exit(-1);
        }
        /* Wait a while before we create the next thread */
        usleep(global_scan_timeout);
    }
    free(ip_addr);

    /* Wait for threads to finish */
    for(i=1; i < 255; i++)
    {
        ret = pthread_join(threads[i-1], NULL);
        if( ret != 0 )
        {
            printf("Error: Return code from pthread_join is: %d for thread: %d\n", ret, i-1);
        }
    }
}


/* Returns true for local subnets to be scanned */
int is_local_network(char *net_addr)
{
    char *data, *tmp;
    int i = 0, ret = 0;

    if( net_addr == NULL || strlen(net_addr) < 3 )
        return 0;

    data = allocate(1024);
    snprintf(data, 1000, "%s", net_addr);

    /* Cut at the last dot */
    for(i=strlen(data)-1; data[i]!='\0'; i--)
    if( data[i]=='.' )
    {
        data[i]='\0';
        break;
    }

    /* Dont scan loopback network */
    if( strcmp(data, "127.0") == 0 )
    {
        printf("Skipping loopback network: %s\n", net_addr);
        free(data);
        return 0;
    }

    /* Dont scan interfaces without adresses */
    if( strcmp(data, "0.0.0.0") == 0 )
    {
        printf("Skipping interface without a valid IPv4 address: %s\n", net_addr);
        free(data);
        return 0;
    }

    /* Valid networks to scan */
    if( strcmp(data, "192.168") == 0
    ||  strcmp(data, "169.254") == 0 )
    {
        free(data);
        return 1;
    }

    /* Valid local net from 172.16 to 172.31 */
    tmp = allocate(1024);

    for(i=16; i<32; i++)
    {
        snprintf(tmp, 1000, "172.%d", i);
        if( strcmp(tmp, data) == 0 )
        {
            ret = 1;
            break;
        }
    }

    /* Valid local net from 10.0 to 10.255 */
    for(i=0; i<256; i++)
    {
        snprintf(tmp, 1000, "10.%d", i);
        if( strcmp(tmp, data) == 0 )
        {
            ret = 1;
            break;
        }
    }

    free(tmp);
    free(data);

    return ret;
}


void handle_arguments(int argc, char *argv[])
{
    int i = 0, z = 0;
    int ch = 0;

    for(i=1; i <= argc; i++)
    {
        /* Verbosity and help options */
        if( argv[i] != NULL )
        {
            if( strcmp(argv[i], "--verbose") == 0 )
                global_verbose = 1;

            if( strcmp(argv[i], "--help") == 0 )
            {
                show_help();
                exit(1);
            }
        }
        /* Set new scan timeout if specified */
        if( argv[i] == NULL || argv[i+1] == NULL )
            continue;

        if( strcmp(argv[i], "--timeout") != 0 )
            continue;

        for(z=0; z < strlen(argv[i+1]); z++)
        {
            /* Max scan timeout is 999999 */
            if( z > 5 )
            {
                printf("Scan timeout max is 999999\n");
                exit(1);
            }
            ch = argv[i+1][z];
            ch = isdigit(ch);
            if( ch == 0 )
            {
                printf("Scan timeout can only contain digits.\n");
                exit(1);
            }
        }

        if( atoi(argv[i+1]) > 9000 - 1 )
        {
            global_scan_timeout = atoi(argv[i+1]);
            printf("Scan timeout set to: %d\n", global_scan_timeout);
        }
        else
        {
            printf("Scan timeout value is too low.\n");
            printf("Minimum timeout is 9000.\n");
            exit(1);
        }
    }
}


int main(int argc, char *argv[])
{
    int sd;
    struct ifreq *ifr, *ifend;
    struct ifreq ifreq;
    struct ifconf ifc;
    struct ifreq ifs[MAX_INTERFACES];
    char *netaddr;

    if( ! nmblookup_exists() )
    {
        printf("Error: Missing nmblookup command.\n");
        exit(-1);
    }


    if( argc > 0 )
        handle_arguments(argc, argv);

    sd = socket(AF_INET, SOCK_DGRAM, 0);

    ifc.ifc_len = sizeof(ifs);
    ifc.ifc_req = ifs;

    if( ioctl(sd, SIOCGIFCONF, &ifc) < 0 )
    {
        printf("ioctl(SIOCGIFCONF)\n");
        close(sd);
        return 0;
    }

    ifend = ifs + (ifc.ifc_len / sizeof(struct ifreq));

    for(ifr = ifc.ifc_req; ifr < ifend; ifr++)
    {
        if( ifr->ifr_addr.sa_family == AF_INET )
        {
            strncpy(ifreq.ifr_name, ifr->ifr_name, sizeof(ifreq.ifr_name));

            printf("\nNetwork device:\t %s\n", ifreq.ifr_name);

            /* Print IPv4 address */
            if( ioctl(sd, SIOCGIFADDR, &ifreq) < 0 )
            {
                printf("SIOCGIFADDR(%s): %m\n", ifreq.ifr_name);
                close(sd);
                return 0;
            }

            netaddr = allocate(1024);
            snprintf(netaddr, 1000, "%lu.%lu.%lu",
            (long unsigned int) ((unsigned char *) &ifreq.ifr_addr.sa_data)[2],
            (long unsigned int) ((unsigned char *) &ifreq.ifr_addr.sa_data)[3],
            (long unsigned int) ((unsigned char *) &ifreq.ifr_addr.sa_data)[4]);

            /* Only scan local networks */
            if( ! is_local_network(netaddr) )
            {
                free(netaddr);
                continue;
            }

            scan_samba_hosts(netaddr);
            free(netaddr);
        }
    }

    close(sd);
    return 0;
}

